"""The Level-Based Foraging Environment.

This implementation uses and is based on the original implementation of the
Level-Based Foraging environment:

https://github.com/semitable/lb-foraging

References
----------
- Stefano V. Albrecht and Subramanian Ramamoorthy. 2013. A Game-Theoretic Model and
  Best-Response Learning Method for Ad Hoc Coordination in Multia-gent Systems.
  In Proceedings of the 2013 International Conference on Autonomous Agents and
  Multi-Agent Systems. 1155–1156.
- S. V. Albrecht and Peter Stone. 2017. Reasoning about Hypothetical Agent
  Behaviours and Their Parameters. In 16th International Conference on Autonomous
  Agents and Multiagent Systems 2017. International Foundation for Autonomous
  Agents and Multiagent Systems, 547–555
- Filippos Christianos, Lukas Schäfer, and Stefano Albrecht. 2020. Shared Experience
  Actor-Critic for Multi-Agent Reinforcement Learning. Advances in Neural
  Information Processing Systems 33 (2020), 10707–10717
- Georgios Papoudakis, Filippos Christianos, Lukas Schäfer, and Stefano V. Albrecht.
  2021. Benchmarking Multi-Agent Deep Reinforcement Learning Algorithms in
  Cooperative Tasks. In Thirty-Fifth Conference on Neural Information Processing
  Systems Datasets and Benchmarks Track (Round 1)

"""
import enum
import math
from collections import defaultdict
from itertools import product
from os import path
from typing import Dict, List, NamedTuple, Optional, SupportsFloat, Tuple, Union

import numpy as np
from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.error import DependencyNotInstalled
from posggym.utils import seeding


class Player(NamedTuple):
    """A player in the Level-Based Foraging environment."""

    idx: int
    coord: Coord
    level: int


class Food(NamedTuple):
    """A food in the Level-Based Foraging environment."""

    coord: Coord
    level: int


class LBFState(NamedTuple):
    """State in Level-Based Foraging environment."""

    players: Tuple[Player, ...]
    food: Tuple[Food, ...]
    food_spawned: int


class LBFAction(enum.IntEnum):
    """Level-Based Foraging Actions."""

    NONE = 0
    NORTH = 1
    SOUTH = 2
    WEST = 3
    EAST = 4
    LOAD = 5


ACTION_TO_DIR = {
    LBFAction.NORTH: Direction.NORTH,
    LBFAction.SOUTH: Direction.SOUTH,
    LBFAction.WEST: Direction.WEST,
    LBFAction.EAST: Direction.EAST,
}


class CellEntity(enum.IntEnum):
    """Entity encodings for grid observations."""

    OUT_OF_BOUNDS = 0
    EMPTY = 1
    FOOD = 2
    AGENT = 3


LBFObs = Union[Tuple[int, ...], np.ndarray]


class LBFEntityObs(NamedTuple):
    """Entity (player, food) observation in the LBF environment."""

    coord: Coord
    level: int
    is_self: bool


class LBFEnv(DefaultEnv[LBFState, LBFObs, LBFAction]):
    """The Level-Based Foraging Environment.

    This implementation uses and is based on the original implementation of
    the Level-Based Foraging environment:

    https://github.com/semitable/lb-foraging

    We provide a discription here for convinience.

    Agents
    ------
    Varied number

    State
    -----
    The state of the environment is defined by a (x, y, level) triplet for each
    agent and food object in the environment. The (x, y) components define the
    position, starting from the bottom left square, while the level is the
    level of the agent or food.

    Actions
    -------
    Each agent has six possible discrete actions [0-5] corresponding to:

        Noop, Move North, Move South, Move West, Move East, Pickup

    Observations
    ------------
    Each agent observes the (x, y, level) of food and other agent within their
    field of vision, which is the grid `sight` distance away from the agent in
    all directions.

    There are three observation modes:

    1. grid_observation
       The agent recieves three 2D layers of size (1+2*`sight`, 1+2*`sight`).
       Each cell in each layer corresponds to a specific (x, y) coordinate
       relative to the observing agent.
       The layers are:
         i. agent level
         ii. food level
         iii. whether cell is free or blocked (e.g. out of bounds)
    2. vector observation
       A vector of (x, y, level) triplets for each food and agent in the
       environment. If a given food or agent is not within the observing
       agent's field of vision triplets have a value of [-1, -1, 0].
       The size of the vector is (`num_agents` + `max_food`) * 3, with the
       first `max_food` triplets being for the food, the `max_food`+1 triplet
       being for the observing agent and the remaining `num_agents`-1
       triplets for the other agents.
       The ordering of the triplets for the other agents is consistent, while
       the food obs triplets can change based on how many are visible and their
       relative coordinates to the observing agent.
    3. tuple observation
       This is the same as the vector observation except observations are
       Python tuples of integers instead of numpy arrays of floats.

    References
    ----------
    - Stefano V. Albrecht and Subramanian Ramamoorthy. 2013. A Game-Theoretic Model and
      Best-Response Learning Method for Ad Hoc Coordination in Multia-gent Systems.
      In Proceedings of the 2013 International Conference on Autonomous Agents and
      Multi-Agent Systems. 1155–1156.
    - S. V. Albrecht and Peter Stone. 2017. Reasoning about Hypothetical Agent
      Behaviours and Their Parameters. In 16th International Conference on Autonomous
      Agents and Multiagent Systems 2017. International Foundation for Autonomous
      Agents and Multiagent Systems, 547–555
    - Filippos Christianos, Lukas Schäfer, and Stefano Albrecht. 2020. Shared Experience
      Actor-Critic for Multi-Agent Reinforcement Learning. Advances in Neural
      Information Processing Systems 33 (2020), 10707–10717
    - Georgios Papoudakis, Filippos Christianos, Lukas Schäfer, and Stefano V. Albrecht.
      2021. Benchmarking Multi-Agent Deep Reinforcement Learning Algorithms in
      Cooperative Tasks. In Thirty-Fifth Conference on Neural Information Processing
      Systems Datasets and Benchmarks Track (Round 1)

    """

    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 10}

    def __init__(
        self,
        num_agents: int,
        max_agent_level: int,
        field_size: Tuple[int, int],
        max_food: int,
        sight: int,
        force_coop: bool,
        static_layout: bool,
        normalize_reward: bool = True,
        observation_mode: str = "tuple",
        penalty: float = 0.0,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            LBFModel(
                num_agents,
                max_agent_level,
                field_size=field_size,
                max_food=max_food,
                sight=sight,
                force_coop=force_coop,
                static_layout=static_layout,
                normalize_reward=normalize_reward,
                observation_mode=observation_mode,
                penalty=penalty,
                **kwargs,
            ),
            render_mode=render_mode,
        )

        grid: Grid = self.model.grid  # type: ignore
        self.window_size = (min(64 * grid.width, 512), min(64 * grid.height, 512))
        self.cell_size = (
            self.window_size[0] // grid.width,
            self.window_size[1] // grid.height,
        )
        self.window_surface = None
        self.clock = None
        self.food_img = None
        self.agent_img = None
        self.font = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        try:
            import pygame
        except ImportError as e:
            raise DependencyNotInstalled(
                "pygame is not installed, run `pip install posggym[grid-world]`"
            ) from e

        if self.window_surface is None:
            pygame.init()

            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption("Level-Based Foraging")
                self.window_surface = pygame.display.set_mode(self.window_size)
            elif self.render_mode == "rgb_array":
                pygame.font.init()
                self.window_surface = pygame.Surface(self.window_size)

        assert (
            self.window_surface is not None
        ), "Something went wrong with pygame. This should never happen."

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.agent_img is None:
            file_name = path.join(path.dirname(__file__), "img/robot.png")
            self.agent_img = pygame.transform.scale(
                pygame.image.load(file_name), self.cell_size
            )
        if self.food_img is None:
            file_name = path.join(path.dirname(__file__), "img/apple.png")
            self.food_img = pygame.transform.scale(
                pygame.image.load(file_name), self.cell_size
            )
        if self.font is None:
            self.font = pygame.font.SysFont("Comic Sans MS", self.cell_size[1] // 4)

        self.window_surface.fill((0, 0, 0))
        grid: Grid = self.model.grid  # type: ignore

        # draw grid lines
        for y in range(grid.height):
            for x in range(grid.width):
                pos = (x * self.cell_size[0], y * self.cell_size[1])
                rect = (*pos, *self.cell_size)
                pygame.draw.rect(self.window_surface, (255, 255, 255), rect, 1)

        for player in self._state.players:
            (x, y) = player.coord
            pos = (x * self.cell_size[0], y * self.cell_size[1])
            self.window_surface.blit(self.agent_img, pos)
            level_txt = self.font.render(
                str(player.level), True, (255, 255, 255), (0, 0, 0)
            )
            text_pos = (pos[0] + 1, pos[1] + 1)
            self.window_surface.blit(level_txt, text_pos)

        for food in self._state.food:
            (x, y) = food.coord
            pos = (x * self.cell_size[0], y * self.cell_size[1])
            self.window_surface.blit(self.food_img, pos)
            level_txt = self.font.render(str(food.level), True, (255, 255, 255))
            text_pos = (pos[0] + 1, pos[1] + 1)
            self.window_surface.blit(level_txt, text_pos)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
        elif self.render_mode == "rgb_array":
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
            )

    def close(self):
        if self.window_surface is not None:
            import pygame

            pygame.display.quit()
            pygame.quit()


class LBFModel(M.POSGModel[LBFState, LBFObs, LBFAction]):
    """Level-Based Foraging model.

    Parameters
    ----------
    num_agents : int
        the number of agents in the environment
    max_agent_level : int
        maximum foraging level of any agent
    field_size : Tuple[int, int]
        width and height of the playing field
    max_food : int
        the max number of food that can be spawned
    sight : int
        observation size of each agent
    force_coop : bool
        whether to force cooperation or not
    normalize_reward : bool, optional
        whether to normalize the reward or not (default=True)
    observation_mode : str, optional
        The observation mode for agent (default='tuple')
          - 'grid' - observations are multiple 2D grids (3D np.ndarray)
          - 'vector' - observations are vector (1D np.ndarray)
          - 'tuple' - observations are a tuple with same format as 'vector'
                      observations but as a hashable Python tuple object
                      containing integers instead of floats

    penalty : float, optional
        the penalty for failing to load food (default=0.0)

    """

    OBSERVATION_MODES = ["grid", "vector", "tuple"]

    # pylint: disable=unused-argument
    def __init__(
        self,
        num_agents: int,
        max_agent_level: int,
        field_size: Tuple[int, int],
        max_food: int,
        sight: int,
        force_coop: bool,
        static_layout: bool,
        normalize_reward: bool = True,
        observation_mode: str = "tuple",
        penalty: float = 0.0,
        **kwargs,
    ):
        assert observation_mode in self.OBSERVATION_MODES
        self._field_size = field_size
        self._rows, self._cols = field_size
        self._max_agent_level = max_agent_level
        self._max_food = max_food
        self._sight = sight
        self._force_coop = force_coop
        self._static_layout = static_layout
        self._normalize_reward = normalize_reward
        self._obs_mode = observation_mode
        self._penalty = penalty

        self.possible_agents = tuple(range(num_agents))
        self.action_spaces = {
            i: spaces.Discrete(len(LBFAction)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: self.get_agent_observation_space() for i in self.possible_agents
        }
        self.observation_first = True
        self.is_symmetric = True

        self.grid = Grid(
            grid_width=self._cols, grid_height=self._rows, block_coords=None
        )

        self._food_locations: Optional[List[Coord]] = None
        self._player_locations: Optional[List[Coord]] = None
        if static_layout:
            assert (
                self._rows == self._cols
            ), "static layout only supported for square grids"
            self._food_locations = self._generate_static_food_coords()
            self._player_locations = self._generate_static_player_coords()

    def get_agent_observation_space(self) -> spaces.Space:
        """The Observation Space for an agent.

        - all of the board (board_size^2) with foods
        - player description (x, y, level)*player_count

        """
        n_agents = len(self.possible_agents)
        max_food_level = self._max_agent_level * min(n_agents, 4)
        if self._obs_mode == "tuple":
            max_x, max_y = self._cols, self._rows
            tuple_obs_space = []
            for _ in range(n_agents):
                tuple_obs_space.append(spaces.Discrete(max_x + 1, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_y + 1, start=-1))
                tuple_obs_space.append(
                    spaces.Discrete(self._max_agent_level + 1, start=0)
                )
            for _ in range(self._max_food):
                tuple_obs_space.append(spaces.Discrete(max_x + 2, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_y + 2, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_food_level + 1, start=0))
            return spaces.Tuple(tuple_obs_space)

        if self._obs_mode == "vector":
            max_x, max_y = self._cols - 1, self._rows - 1
            min_obs = [-1, -1, 0] * n_agents + [-1, -1, 0] * self._max_food
            max_obs = [max_x, max_y, self._max_agent_level] * n_agents + [
                max_x,
                max_y,
                max_food_level,
            ] * self._max_food
        else:
            # grid observation space
            grid_shape = (1 + 2 * self._sight, 1 + 2 * self._sight)

            # agents layer: agent levels
            agents_min = np.zeros(grid_shape, dtype=np.float32)
            agents_max = np.ones(grid_shape, dtype=np.float32) * self._max_agent_level

            # foods layer: foods level
            foods_min = np.zeros(grid_shape, dtype=np.float32)
            foods_max = np.ones(grid_shape, dtype=np.float32) * max_food_level

            # access layer: i the cell available
            access_min = np.zeros(grid_shape, dtype=np.float32)
            access_max = np.ones(grid_shape, dtype=np.float32)

            # total layer
            min_obs = np.stack([agents_min, foods_min, access_min])
            max_obs = np.stack([agents_max, foods_max, access_max])

        # dtype = np.int8 if self._obs_mode == "vector" else np.uint8
        dtype = np.float32
        return spaces.Box(
            np.array(min_obs, dtype=dtype),
            np.array(max_obs, dtype=dtype),
            dtype=dtype,
        )

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        if self._normalize_reward:
            reward_range = (self._penalty, 1.0)
        else:
            reward_range = (self._penalty, self._max_agent_level)
        return {i: reward_range for i in self.possible_agents}

    def get_agents(self, state: LBFState) -> List[M.AgentID]:
        return list(self.possible_agents)

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def seed(self, seed: Optional[int] = None):
        super().seed(seed)
        if self._static_layout:
            assert (
                self._rows == self._cols
            ), "static layout only supported for square grids"
            self._food_locations = self._generate_static_food_coords()
            self._player_locations = self._generate_static_player_coords()

    def _generate_static_food_coords(self) -> List[Coord]:
        """Generate food coords for static layout.

        Number and location of food is the same for given pairing of
        (field size, max_food).
        """
        assert self._rows == self._cols
        # must be enough space to surround each food with free cells
        assert self._max_food <= math.floor((self._rows - 1) / 2) ** 2

        food_grid_size = math.floor((self._rows - 1) / 2)
        available_food_locs = sorted_from_middle(
            list(product(range(food_grid_size), repeat=2))
        )

        food_locs = []
        for f in range(self._max_food):
            row = available_food_locs[f][0] * 2 + 1
            col = available_food_locs[f][1] * 2 + 1
            food_locs.append((row, col))
        return food_locs

    def _generate_static_player_coords(self) -> List[Coord]:
        """Generate player start coords for static layout.

        Players always start around edge of field.
        """
        assert self._rows == self._cols

        assert len(self.possible_agents) <= math.ceil(self._rows / 2) * 4 - 4

        idxs = [
            i * 2 if i < self._rows / 4 else i * 2 + ((self._rows + 1) % 2)
            for i in range(math.ceil(self._rows / 2))
        ]
        idxs_reverse = list(reversed(idxs))

        # set it so locations are chosen from corners first then
        # alternating sides
        sides = [
            product([0], idxs[:-1]),
            product([self._rows - 1], idxs_reverse[:-1]),
            product(idxs_reverse[:-1], [0]),
            product(idxs[:-1], [self._rows - 1]),
        ]
        available_locations: List[Coord] = []
        for locs in zip(*sides):
            available_locations.extend(locs)

        return available_locations[: len(self.possible_agents)]

    def sample_initial_state(self) -> LBFState:
        if self._static_layout:
            players = self._spawn_players_static()
        else:
            players = self._spawn_players_generative()

        player_levels = sorted([player.level for player in players])
        max_level = sum(player_levels[:3])

        if self._static_layout:
            food = self._spawn_food_static(max_level)
        else:
            food = self._spawn_food_generative(
                self._max_food, max_level, [p.coord for p in players]
            )
        return LBFState(players, food, sum(f.level for f in food))

    def _spawn_players_static(self) -> Tuple[Player, ...]:
        assert self._player_locations is not None
        players = []
        for i in range(len(self.possible_agents)):
            (x, y) = self._player_locations[i]
            level = self.rng.randint(1, self._max_agent_level - 1)  # type: ignore
            players.append(Player(i, (x, y), level))
        return tuple(players)

    def _spawn_players_generative(self) -> Tuple[Player, ...]:
        players = []
        available_coords = list(product(range(self._cols), range(self._rows)))
        for i in range(len(self.possible_agents)):
            level = self.rng.randint(1, self._max_agent_level - 1)  # type: ignore
            coord = self.rng.choice(available_coords)
            available_coords.remove(coord)
            players.append(Player(i, coord, level))
        return tuple(players)

    def _spawn_food_static(self, max_level: int) -> Tuple[Food, ...]:
        """Spawn food in static layout.

        Number and location of food is the same for given pairing of
        (field size, max_food), only the levels of each food will change.
        """
        assert self._food_locations is not None
        assert max_level >= 1
        min_level = max_level if self._force_coop else 1
        food = []
        for (x, y) in self._food_locations:
            level = min_level
            if min_level != max_level:
                level = self.rng.randint(min_level, max_level)  # type: ignore
            food.append(Food((x, y), level))
        return tuple(food)

    def _spawn_food_generative(
        self,
        max_food: int,
        max_level: int,
        player_coords: List[Coord],
    ) -> Tuple[Food, ...]:
        assert max_level >= 1
        attempts = 0
        min_level = max_level if self._force_coop else 1
        unavailable_coords = set(player_coords)
        food: List[Food] = []
        while len(food) < max_food and attempts < 1000:
            attempts += 1
            x = self.rng.randint(1, self._cols - 2)  # type: ignore
            y = self.rng.randint(1, self._rows - 2)  # type: ignore

            if (x, y) in unavailable_coords:
                continue
            # add coord and adjacent coord to unavailable set so food are always
            # atleast one cell apart
            for x_delta, y_delta in product([-1, 0, 1], repeat=2):
                unavailable_coords.add((x + x_delta, y + y_delta))
            level = min_level
            if min_level != max_level:
                level = self.rng.randint(min_level, max_level)  # type: ignore
            food.append(Food((x, y), level))
        return tuple(food)

    def sample_initial_obs(self, state: LBFState) -> Dict[M.AgentID, LBFObs]:
        return self._get_obs(state)

    def step(
        self, state: LBFState, actions: Dict[M.AgentID, LBFAction]
    ) -> M.JointTimestep[LBFState, LBFObs]:
        assert all(0 <= a < len(LBFAction) for a in actions.values())
        next_state, rewards = self._get_next_state(state, actions)
        obs = self._get_obs(next_state)
        all_done = len(next_state.food) == 0
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: LBFState, actions: Dict[M.AgentID, LBFAction]
    ) -> Tuple[LBFState, Dict[M.AgentID, SupportsFloat]]:
        next_food = {f.coord: f for f in state.food}

        # try move agents
        # if two or more agents try to move to the same location they all fail
        collisions = defaultdict(list)
        loading_players = set()
        for player in state.players:
            i = player.idx
            a_i = actions[i]
            if a_i == LBFAction.NONE:
                next_coord = player.coord
            elif a_i == LBFAction.LOAD:
                next_coord = player.coord
                loading_players.add(player)
            else:
                next_coord = self.grid.get_next_coord(player.coord, ACTION_TO_DIR[a_i])
            if next_coord in next_food:
                next_coord = player.coord
            collisions[next_coord].append(player)

        # do movements for non colliding players
        next_players = list(state.players)
        for coord, player_list in collisions.items():
            # make sure no more than an player will arrive at location
            if len(player_list) > 1:
                continue
            player = player_list[0]
            next_players[player.idx] = Player(player.idx, coord, player.level)

        # process the loadings and calculate rewards as necessary
        rewards: Dict[M.AgentID, float] = {i: 0.0 for i in self.possible_agents}
        for player in loading_players:
            adj_coords = self.grid.get_neighbours(player.coord)
            for adj_coord in adj_coords:
                if adj_coord not in next_food:
                    continue
                adj_food = next_food[adj_coord]
                food_adj_coords = self.grid.get_neighbours(adj_coord)
                adj_players = [p for p in loading_players if p.coord in food_adj_coords]
                adj_player_level = sum([p.level for p in adj_players])

                if adj_player_level < adj_food.level:
                    # failed to load
                    for p in adj_players:
                        rewards[p.idx] -= self._penalty
                else:
                    # food was loaded and each player scores points
                    for p in adj_players:
                        p_reward = float(p.level * adj_food.level)
                        if self._normalize_reward:
                            p_reward = p_reward / float(
                                adj_player_level * state.food_spawned
                            )
                        rewards[p.idx] += p_reward
                    # and food is removed
                    next_food.pop(adj_coord)
        next_state = LBFState(
            tuple(next_players), tuple(next_food.values()), state.food_spawned
        )
        return next_state, rewards  # type: ignore

    def _get_obs(self, state: LBFState) -> Dict[M.AgentID, LBFObs]:
        obs: Dict[M.AgentID, LBFObs] = {}
        for i in self.possible_agents:
            player_obs, food_obs = self._get_local_obs(state, int(i))
            if self._obs_mode == "tuple":
                obs[i] = self._get_tuple_obs(player_obs, food_obs)
            elif self._obs_mode == "vector":
                obs[i] = self._get_vector_obs(int(i), player_obs, food_obs)
            else:
                obs[i] = self._get_grid_obs(
                    int(i), state.players[int(i)].coord, player_obs, food_obs
                )
        return obs

    def _get_local_obs(
        self, state: LBFState, agent_id: int
    ) -> Tuple[List[LBFEntityObs], List[LBFEntityObs]]:
        # player is always in center of observable area
        player_obs = []
        ego_player = state.players[agent_id]
        for p in state.players:
            if (
                abs(p.coord[0] - ego_player.coord[0]) <= self._sight
                and abs(p.coord[1] - ego_player.coord[1]) <= self._sight
            ):
                p_obs_coord = (
                    p.coord[0] - ego_player.coord[0] + self._sight,
                    p.coord[1] - ego_player.coord[1] + self._sight,
                )
                player_obs.append(
                    LBFEntityObs(p_obs_coord, p.level, is_self=(ego_player == p))
                )
            else:
                # player p not observed
                player_obs.append(LBFEntityObs((-1, -1), 0, is_self=False))

        food_obs = []
        for f in state.food:
            # note unobserved food are exluded since ordering can change for food obs
            # (unlike for players)
            if (
                abs(f.coord[0] - ego_player.coord[0]) <= self._sight
                and abs(f.coord[1] - ego_player.coord[1]) <= self._sight
            ):
                f_obs_coord = (
                    f.coord[0] - ego_player.coord[0] + self._sight,
                    f.coord[1] - ego_player.coord[1] + self._sight,
                )
                food_obs.append(LBFEntityObs(f_obs_coord, f.level, is_self=False))

        return player_obs, food_obs

    def _get_tuple_obs(
        self, player_obs: List[LBFEntityObs], food_obs: List[LBFEntityObs]
    ) -> LBFObs:
        obs: List[int] = []
        for o in player_obs:
            if o.is_self:
                obs.insert(0, o.level)
                obs.insert(0, o.coord[1])
                obs.insert(0, o.coord[0])
            else:
                obs.extend((*o.coord, o.level))
        for f_idx in range(self._max_food):
            if f_idx < len(food_obs):
                obs.extend((*food_obs[f_idx].coord, food_obs[f_idx].level))
            else:
                obs.extend((-1, -1, 0))
        return tuple(obs)

    def _get_vector_obs(
        self,
        agent_id: int,
        player_obs: List[LBFEntityObs],
        food_obs: List[LBFEntityObs],
    ) -> np.ndarray:
        # initialize obs array to (-1, -1, 0)
        obs = np.full(
            ((len(self.possible_agents) + self._max_food) * 3,), -1.0, dtype=np.float32
        )
        obs[2::3] = 0
        for i, o in enumerate(player_obs):
            if o.level == 0:
                # player unobserved
                continue
            if o.is_self:
                p_idx = 0
            else:
                p_idx = i * 3 if i > agent_id else (i + 1) * 3
            obs[p_idx : p_idx + 3] = (*o.coord, o.level)

        food_start_idx = len(self.possible_agents) * 3
        for i in range(self._max_food):
            if i < len(food_obs):
                f_idx = food_start_idx + (i * 3)
                obs[f_idx : f_idx + 3] = (*food_obs[i].coord, food_obs[i].level)

        return obs

    def _get_grid_obs(
        self,
        agent_id: int,
        agent_coord: Coord,
        player_obs: List[LBFEntityObs],
        food_obs: List[LBFEntityObs],
    ) -> np.ndarray:
        grid_shape_x, grid_shape_y = (2 * self._sight + 1, 2 * self._sight + 1)
        # agent, food, access layers
        grid_obs = np.zeros((3, grid_shape_x, grid_shape_y), dtype=np.float32)

        # add all in-bounds areas as accessible
        # cells containing players and food will be made inaccessible in next steps
        agent_x, agent_y = agent_coord
        for x, y in product(
            range(
                max(agent_x - self._sight, 0), min(self._cols, agent_x + self._sight)
            ),
            range(
                max(agent_y - self._sight, 0), min(self._rows, agent_y + self._sight)
            ),
        ):
            obs_x = x - agent_x + self._sight
            obs_y = y - agent_y + self._sight
            grid_obs[2, obs_x, obs_y] = 1.0

        for p in player_obs:
            if p.level > 0:
                grid_obs[0][p.coord] = p.level
                # agent coords are not accessible
                grid_obs[2][p.coord] = 0.0

        for f in food_obs:
            grid_obs[1][f.coord] = f.level
            # food coords are not accessible
            grid_obs[2][f.coord] = 0.0

        return grid_obs

    def parse_obs(
        self, obs: LBFObs
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse observation into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        if self._obs_mode == "tuple":
            assert isinstance(obs, tuple)
            return self.parse_tuple_obs(obs)

        assert isinstance(obs, np.ndarray)
        if self._obs_mode == "grid":
            return self.parse_grid_obs(obs)
        # vector obs
        return self.parse_vector_obs(obs)

    def parse_grid_obs(
        self, obs: np.ndarray
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse grid observation int (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        raise NotImplementedError

    def parse_vector_obs(
        self, obs: np.ndarray
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse vector obs into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        assert obs.shape[0] == 3 * (len(self.possible_agents) + self._max_food)
        agent_obs = []
        food_obs = []
        for i in range(0, obs.shape[0], 3):
            triplet = (int(obs[i]), int(obs[i + 1]), int(obs[i + 2]))
            if i < len(self.possible_agents) * 3:
                agent_obs.append(triplet)
            else:
                food_obs.append(triplet)
        return agent_obs, food_obs

    def parse_tuple_obs(
        self, obs: Tuple[int, ...]
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse tuple obs into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        assert len(obs) == 3 * (len(self.possible_agents) + self._max_food)
        agent_obs = []
        food_obs = []
        for i in range(0, len(obs), 3):
            triplet = (int(obs[i]), int(obs[i + 1]), int(obs[i + 2]))
            if i < len(self.possible_agents) * 3:
                agent_obs.append(triplet)
            else:
                food_obs.append(triplet)
        return agent_obs, food_obs


def sorted_from_middle(lst):
    """Sorts list from middle out."""
    left = lst[len(lst) // 2 - 1 :: -1]
    right = lst[len(lst) // 2 :]
    output = [right.pop(0)] if len(lst) % 2 else []
    for t in zip(left, right):
        output += sorted(t)
    return output



"""A set of functions for passively checking environment implementations.

Ref:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/utils/passive_env_checker.py

"""
import inspect
import random
from functools import partial
from typing import Callable, Dict, Optional, Sequence, get_args

import gymnasium.utils.passive_env_checker as gym_passive_env_checker
import numpy as np
from gymnasium import Space, spaces
from gymnasium.utils.env_checker import check_space_limit

import posggym
import posggym.model as M
from posggym import error, logger
from posggym.utils import seeding


def data_equivalence(data_1, data_2) -> bool:
    """Assert equality between data 1 and 2, i.e observations, actions, info.

    Arguments
    ---------
    data_1: data structure 1
    data_2: data structure 2

    Returns
    -------
    bool: If observation 1 and 2 are equivalent

    """
    if type(data_1) == type(data_2):
        if isinstance(data_1, dict):
            return data_1.keys() == data_2.keys() and all(
                data_equivalence(data_1[k], data_2[k]) for k in data_1.keys()
            )
        elif isinstance(data_1, (tuple, list)):
            return len(data_1) == len(data_2) and all(
                data_equivalence(o_1, o_2) for o_1, o_2 in zip(data_1, data_2)
            )
        elif isinstance(data_1, np.ndarray):
            return data_1.shape == data_2.shape and np.allclose(
                data_1, data_2, atol=0.00001
            )
        else:
            return data_1 == data_2
    else:
        return False


def check_rng_equality(rng_1: seeding.RNG, rng_2: seeding.RNG, prefix=None):
    """Check equality between two random number generators."""
    assert type(rng_1) == type(
        rng_2
    ), f"{prefix}Differing RNG types: {rng_1} and {rng_2}"
    if isinstance(rng_1, random.Random) and isinstance(rng_2, random.Random):
        assert (
            rng_1.getstate() == rng_2.getstate()
        ), f"{prefix}Internal states differ: {rng_1} and {rng_2}"
    elif isinstance(rng_1, np.random.Generator) and isinstance(
        rng_2, np.random.Generator
    ):
        assert (
            rng_1.bit_generator.state == rng_2.bit_generator.state
        ), f"{prefix}Internal states differ: {rng_1} and {rng_2}"
    else:
        raise AssertionError(f"{prefix}Unsupported RNG type: '{type(rng_1)}'.")


def check_agent_space_limits(agent_spaces: Dict[M.AgentID, Space], space_type: str):
    """Check the space limit for any agent Box space."""
    for i, agent_space in agent_spaces.items():
        check_space_limit(agent_space, space_type)


def _check_box_state_space(state_space: spaces.Box):
    """Checks that a :class:`Box` state space is defined in a sensible way.

    Arguments
    ---------
    state_space: A box state space.

    """
    assert state_space.low.shape == state_space.shape, (
        "The Box state space shape and low shape have have different shapes, "
        f"low shape: {state_space.low.shape}, box shape: {state_space.shape}"
    )
    assert state_space.high.shape == state_space.shape, (
        f"The Box state space shape and high shape have different shapes, "
        f"high shape: {state_space.high.shape}, box shape: {state_space.shape}"
    )

    if np.any(state_space.low == state_space.high):
        logger.warn(
            "A Box state space maximum and minimum values are equal. "
            "Actual equal coordinates: "
            f"{[x for x in zip(*np.where(state_space.low == state_space.high))]}"
        )
    elif np.any(state_space.high < state_space.low):
        logger.warn(
            "A Box state space low value is greater than a high value. "
            "Actual less than coordinates: "
            f"{[x for x in zip(*np.where(state_space.high < state_space.low))]}"
        )


def check_agent_spaces(
    agent_spaces: Dict[M.AgentID, Space],
    space_type: str,
    check_box_space_fn: Callable[[spaces.Box], None],
):
    """A passive check of environment spaces that should not affect the environment."""
    assert all(isinstance(i, get_args(M.AgentID)) for i in agent_spaces)
    for i, space_i in agent_spaces.items():
        try:
            gym_passive_env_checker.check_space(space_i, space_type, check_box_space_fn)
        except AssertionError as e:
            raise AssertionError(f"Invalid {space_type} space for agent '{i}'.") from e


check_state_space = partial(
    gym_passive_env_checker.check_space,
    space_type="state",
    check_box_space_fn=_check_box_state_space,
)
check_agent_observation_spaces = partial(
    check_agent_spaces,
    space_type="observation",
    check_box_space_fn=gym_passive_env_checker._check_box_observation_space,
)
check_agent_action_spaces = partial(
    check_agent_spaces,
    space_type="action",
    check_box_space_fn=gym_passive_env_checker._check_box_action_space,
)


def check_state(state: M.StateType, model: M.POSGModel):
    """Check state is valid.

    Arguments
    ---------
    state: the state to check
    model: the environment model

    """
    if model.state_space is not None:
        assert model.state_space.contains(state)


def check_agent_obs(
    obs: Dict[M.AgentID, M.ObsType],
    observation_spaces: Dict[M.AgentID, Space],
    method_name: str,
):
    """Check that each agent's observation returned by the environment is valid.

    Arguments
    ---------
    obs: The observation for each agent to check
    observation_spaces: The observation spaces for each agent's observation
    method_name: The method name that generated the observation

    """
    for i, obs_i in obs.items():
        try:
            gym_passive_env_checker.check_obs(obs_i, observation_spaces[i], method_name)
        except AssertionError as e:
            raise AssertionError("Invalid observation for agent `{i}`.") from e


def check_reset_obs(obs: Optional[Dict[M.AgentID, M.ObsType]], model: M.POSGModel):
    """Check agent observations returned by the environment `reset()` method are valid.

    Arguments
    ---------
    obs: The observation for each agent to check
    model: The environment model

    """
    if not model.observation_first:
        assert (
            obs is None
        ), "Expected None from `env.reset()` for `action_first` environment."
        return

    assert isinstance(obs, dict), (
        "Expected observation from `env.reset()` to be a dictionary, mapping agent IDs "
        " to agent obs."
    )
    for i, o_i in obs.items():
        assert (
            i in model.possible_agents
        ), f"Invalid agent ID `{i}`. Possible IDs are {model.possible_agents}."

    check_agent_obs(obs, model.observation_spaces, "reset")


def env_reset_passive_checker(env, **kwargs):
    """A passive check of the `Env.reset` function.

    Investigates the returning reset information and returning the data unchanged.

    """
    signature = inspect.signature(env.reset)
    if "seed" not in signature.parameters and "kwargs" not in signature.parameters:
        logger.warn(
            "posggym requires that `Env.reset` can be passed a `seed` for resetting the"
            " environment random number generator."
        )
    else:
        seed_param = signature.parameters.get("seed")
        # Check the default value is None
        if seed_param is not None and seed_param.default is not None:
            logger.warn(
                "The default seed argument in `Env.reset` should be `None`, otherwise "
                "the environment will by default always be deterministic. "
                f"Actual default: {seed_param}"
            )

    if "options" not in signature.parameters and "kwargs" not in signature.parameters:
        logger.warn(
            "posggym requires that `Env.reset` can be passed `options` to allow the "
            "environment initialisation to be passed additional information."
        )

    # Checks the result of env.reset with kwargs
    result = env.reset(**kwargs)

    if not isinstance(result, tuple):
        logger.warn(
            "The result returned by `env.reset()` was not a tuple of the form "
            "`(obs, info)`, where `obs` is a observation and `info` is a dictionary "
            f"containing additional information. Actual type: `{type(result)}`"
        )
    elif len(result) != 2:
        logger.warn(
            "The result returned by `env.reset()` should be `(obs, info)` by default, "
            "where `obs` is a observation and `info` is a dictionary containing "
            "additional information."
        )
    else:
        obs, info = result
        check_reset_obs(obs, env.model)
        assert isinstance(info, dict), (
            "The second element returned by `env.reset()` was not a dictionary, "
            f"actual type: {type(info)}"
        )
    return result


def _check_agent_dict(
    agent_dict,
    possible_agents: Sequence[M.AgentID],
    dict_type: str,
    expected_agents: Optional[Sequence[M.AgentID]] = None,
):
    assert isinstance(agent_dict, dict), (
        f"Agent {dict_type} dictionary  must be a dictionary mapping agentID to values."
        f"Actual type: {type(agent_dict)}."
    )
    for i in agent_dict:
        assert i in possible_agents, (
            f"Agent {dict_type} dictionary must only contain valid agent IDs: "
            f"invalid ID `{i}`."
        )
    if expected_agents is not None:
        for i in expected_agents:
            assert i in agent_dict, (
                f"Expected agent ID `{i}` missing from {dict_type} dictionary."
            )


def model_step_passive_checker(
    model: M.POSGModel, state: M.StateType, actions: Dict[M.AgentID, M.ActType]
):
    """A passive check for the model step.

    Investigating the returning data then returning the data unchanged.
    """
    # List of agents still active in the state and which we expect to get data for
    step_agents = model.get_agents(state)
    for i in step_agents:
        assert i in model.possible_agents, (
            f"Agent ID not in list of possible IDs. Invalid agent ID '{i}'. "
        )

    # We don't check the actions as out-of-bounds values are allowed in some
    # environments
    result = model.step(state, actions)
    assert isinstance(result, M.JointTimestep), (
        "Expects model.step result to be a `posggym.model.JointTimestep`, "
        f"actual type: {type(result)}"
    )
    next_state, obs, reward, terminated, truncated, done, info = result

    _check_agent_dict(obs, model.possible_agents, "observation", step_agents)
    _check_agent_dict(reward, model.possible_agents, "reward", step_agents)
    _check_agent_dict(terminated, model.possible_agents, "terminated", step_agents)
    _check_agent_dict(truncated, model.possible_agents, "truncated", step_agents)
    # Less strict on checking there are entries for all agents in info values
    # as these are not functionally critical and are more for record keeping
    _check_agent_dict(info, model.possible_agents, "info")

    check_state(next_state, model)
    check_agent_obs(obs, model.observation_spaces, "step")

    if not all(isinstance(t_i, (bool, np.bool_)) for t_i in terminated.values()):
        logger.warn(
            "Expects `terminated` signal to be a boolean for every agent, "
            f"actual types: {list(type(t_i) for t_i in terminated.values())}."
        )
    if not all(isinstance(t_i, (bool, np.bool_)) for t_i in truncated.values()):
        logger.warn(
            "Expects `truncated` signal to be a boolean for every agent, "
            f"actual types: {list(type(t_i) for t_i in truncated.values())}."
        )

    if not isinstance(done, (bool, np.bool_)):
        logger.warn(
            "Expects `done` signal returned by `step()` to be a boolean, "
            f"actual type: {type(info)}"
        )

    if not (
        all(
            np.issubdtype(type(r_i), np.integer)
            or np.issubdtype(type(r_i), np.floating)
            for r_i in reward.values()
        )
    ):
        logger.warn(
            "The reward returned for each agent by `step()` must be a float, int, "
            "np.integer or np.floating, "
            f"actual types: {list(type(r_i) for r_i in reward.values())}."
        )
    else:
        for i, r_i in reward.items():
            if np.isnan(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is a NaN value.")
            if np.isinf(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is an inf value.")

    return result


def env_step_passive_checker(env: posggym.Env, actions: Dict[M.AgentID, M.ActType]):
    """A passive check for the environment step.

    Investigating the returning data then returning the data unchanged.
    """
    # List of agents still active in the state and which we expect to get data for
    step_agents = env.agents
    for i in step_agents:
        assert i in env.possible_agents, (
            f"Agent ID not in list of possible IDs. Invalid agent ID '{i}'. "
        )

    # We don't check the actions as out-of-bounds values are allowed in some
    # environments
    result = env.step(actions)
    assert isinstance(
        result, tuple
    ), f"Expects step result to be a tuple, actual type: {type(result)}"
    if len(result) == 6:
        obs, reward, terminated, truncated, done, info = result

        _check_agent_dict(obs, env.possible_agents, "observation", step_agents)
        _check_agent_dict(reward, env.possible_agents, "reward", step_agents)
        _check_agent_dict(terminated, env.possible_agents, "terminated", step_agents)
        _check_agent_dict(truncated, env.possible_agents, "truncated", step_agents)
        # Less strict on checking entries for all agents in the outcomes
        # and info values as these are not functionally critical and are more for
        # record keeping
        _check_agent_dict(info, env.possible_agents, "info")

        if not all(isinstance(t_i, (bool, np.bool_)) for t_i in terminated.values()):
            logger.warn(
                "Expects `terminated` signal to be a boolean for every agent, "
                f"actual types: {list(type(t_i) for t_i in terminated.values())}."
            )
        if not all(isinstance(t_i, (bool, np.bool_)) for t_i in truncated.values()):
            logger.warn(
                "Expects `truncated` signal to be a boolean for every agent, "
                f"actual types: {list(type(t_i) for t_i in truncated.values())}."
            )

        if not isinstance(done, (bool, np.bool_)):
            logger.warn(
                "Expects `done` signal returned by `step()` to be a boolean, "
                f"actual type: {type(info)}"
            )
    else:
        raise error.Error(
            "Expected `Env.step` to return a six element tuple, actual number of "
            f"elements returned: {len(result)}."
        )

    check_agent_obs(obs, env.observation_spaces, "step")

    if not (
        all(
            np.issubdtype(type(r_i), np.integer)
            or np.issubdtype(type(r_i), np.floating)
            for r_i in reward.values()
        )
    ):
        logger.warn(
            "The reward returned for each agent by `step()` must be a float, int, "
            "np.integer or np.floating, "
            f"actual types: {list(type(r_i) for r_i in reward.values())}."
        )
    else:
        for i, r_i in reward.items():
            if np.isnan(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is a NaN value.")
            if np.isinf(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is an inf value.")

    return result


def _check_render_return(render_mode, render_return):
    """Produces warning if `render_return` doesn't match `render_mode`."""
    gym_passive_env_checker._check_render_return(render_mode, render_return)

    # need to check posggym specific render modes, namely dict renders which
    # return mapping from agentID to agent specific render.
    if render_mode.endswith("_dict"):
        if not isinstance(render_return, dict):
            logger.warn(
                f"Render mode `{render_mode}` should produce a dict, got "
                f"{type(render_return)}"
            )
        else:
            # Check that each item of the dict matches the base render mode
            # Doesn't check dict keys, i.e. agent IDs, are valid
            base_render_mode = render_mode[: -len("_dict")]
            for item in render_return:
                _check_render_return(base_render_mode, item)


def env_render_passive_checker(env: posggym.Env):
    """A passive check of the `Env.render`.

    Checks that render modes/fps in the metadata of the environment are declared.

    This function is exactly the same as
    `gymnasium.utils.passive_env_check.env_render_passive_checker`, we copy it here to
    ensure the updated `_check_render_return` function is used.

    """
    render_modes = env.metadata.get("render_modes")
    if render_modes is None:
        logger.warn(
            "No render modes were declared in the environment "
            "(env.metadata['render_modes'] is None or not defined), you may have "
            "trouble when calling `.render()`."
        )
    else:
        if not isinstance(render_modes, (list, tuple)):
            logger.warn(
                "Expects the render_modes to be a sequence (i.e. list, tuple), "
                f"actual type: {type(render_modes)}"
            )
        elif not all(isinstance(mode, str) for mode in render_modes):
            logger.warn(
                "Expects all render modes to be strings, "
                f"actual types: {[type(mode) for mode in render_modes]}"
            )

        render_fps = env.metadata.get("render_fps")
        # We only require `render_fps` if rendering is actually implemented
        if len(render_modes) > 0:
            if render_fps is None:
                logger.warn(
                    "No render fps was declared in the environment "
                    "(env.metadata['render_fps'] is None or not defined), rendering "
                    "may occur at inconsistent fps."
                )
            else:
                if not (
                    np.issubdtype(type(render_fps), np.integer)
                    or np.issubdtype(type(render_fps), np.floating)
                ):
                    logger.warn(
                        "Expects the `env.metadata['render_fps']` to be an integer "
                        f"or a float, actual type: {type(render_fps)}"
                    )
                else:
                    assert render_fps > 0, (
                        "Expects the `env.metadata['render_fps']` to be greater than "
                        f"zero, actual value: {render_fps}"
                    )

        # env.render is now an attribute with default None
        if len(render_modes) == 0:
            assert env.render_mode is None, (
                "With no render_modes, expects the Env.render_mode to be None, "
                f"actual value: {env.render_mode}"
            )
        else:
            assert env.render_mode is None or env.render_mode in render_modes, (
                "The environment was initialized successfully however with an "
                "unsupported render mode. "
                f"Render mode: {env.render_mode}, modes: {render_modes}"
            )

    result = env.render()    # type: ignore
    if env.render_mode is not None:
        _check_render_return(env.render_mode, result)

    return result



"""The Unmanned Aerial Vehicle Grid World Environment.

An adversarial 2D grid world problem involving two agents, a Unmanned
Aerial Vehicle (UAV) and a fugitive. The UAV's goal is to capture the
fugitive, while the fugitive's goal is to reach the safe house located at
a known fixed location on the grid. The fugitive is considered caught if
it is co-located with the UAV. The UAV observes it's own location and
recieves a noisy observation of the fugitive's location. The fugitive does
not know it's location but it recieves a noisy observation of its relative
direction to the safe house when it is adjacent to the safe house.

Reference
---------
Panella, Alessandro, and Piotr Gmytrasiewicz. 2017. “Interactive POMDPs
with Finite-State Models of Other Agents.” Autonomous Agents and
Multi-Agent Systems 31 (4): 861–904.

"""
from os import path
import itertools
import random
from typing import Dict, List, Optional, Set, SupportsFloat, Tuple, Union

import numpy as np
from gymnasium import spaces

from posggym import logger
import posggym.model as M
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding
from posggym.error import DependencyNotInstalled


UAVState = Tuple[Coord, Coord]
UAVAction = int

# UAV Obs = (uav coord, fug coord)
UAVUAVObs = Tuple[Coord, Coord]
# FUG Obs = house direction
UAVFUGObs = int
UAVObs = Union[UAVUAVObs, UAVFUGObs]

OBSNORTH = 0
OBSSOUTH = 1
OBSLEVEL = 2
OBSNONE = 3
DIR_OBS = [OBSNORTH, OBSSOUTH, OBSLEVEL, OBSNONE]
DIR_OBS_STR = ["N", "S", "L", "0"]


class UAVEnv(DefaultEnv[UAVState, UAVObs, UAVAction]):
    """The Unmanned Aerial Vehicle Grid World Environment.

    An adversarial 2D grid world problem involving two agents, a Unmanned
    Aerial Vehicle (UAV) and a fugitive. The UAV's goal is to capture the
    fugitive, while the fugitive's goal is to reach the safe house located at
    a known fixed location on the grid. The fugitive is considered caught if
    it is co-located with the UAV. The UAV observes it's own location and
    recieves a noisy observation of the fugitive's location. The fugitive does
    not know it's location but it recieves a noisy observation of its relative
    direction to the safe house when it is adjacent to the safe house.

    Agents
    ------
    UAV = 0
    Fugitive = 1

    State
    -----
    Each state contains the (x, y) (x=column, y=row, with origin at the
    top-left square of the grid) of the UAV and fugitive agent. Specifically,
    a states is ((x_uav, y_uav), (x_fugitive, y_fugitive))

    Initially, the location of both agents is chosen at random.

    Actions
    -------
    Each agent has 4 actions corresponding to moving in the 4 cardinal
    directions (NORTH=0, EAST=1, SOUTH=2, WEST=3).

    Observation
    -----------
    The UAV observes its (x, y) coordinates and recieves a noisy observation
    of the fugitives (x, y) coordinates. The UAV observes the correct fugitive
    coordinates with p=0.9, and one of the adjacent locations to the true
    fugitive location with p=1-0.9.

    The fugitive can sense it's position with respect to the safe house, namely
    whether it is north of it (OBSNORTH=0), south of it (OBSSOUTH=1), or at the
    same level (OBSLEVEL=3). These observations are recieved with accuracy 0.8,
    and only when the fugitive is adjacent to it. If the fugitive is not
    adjacent to the safe house it recieves no observation (OBSNONE=4).

    Reward
    ------
    Both agents receive a penalty of -0.04 for each step.
    If the fugitive reaches the safe house then the fugitive recieves a reward
    of 1, while the UAV recieves a penalty of -1.
    If the fugitive is caught by the UAV, then the fugitive recieves a penalty
    of -1, while the UAV recieves a reward of 1.

    Transition Dynamics
    -------------------
    Actions are deterministic. The fugitive's position is reset at random if
    it reaches the safe house or gets caught by the UAV.

    Reference
    ---------
    Panella, Alessandro, and Piotr Gmytrasiewicz. 2017. “Interactive POMDPs
    with Finite-State Models of Other Agents.” Autonomous Agents and
    Multi-Agent Systems 31 (4): 861–904.
    """

    metadata = {"render_modes": ["human", "ansi", "rgb_array"], "render_fps": 4}

    def __init__(self, grid_name: str, render_mode: Optional[str] = None, **kwargs):
        super().__init__(UAVModel(grid_name, **kwargs), render_mode=render_mode)

        grid: UAVGrid = self.model.grid   # type: ignore
        self.window_size = (min(64 * grid.width, 512), min(64 * grid.height, 512))
        self.cell_size = (
            self.window_size[0] // grid.width,
            self.window_size[1] // grid.height
        )
        self.window_surface = None
        self.clock = None
        self.uav_img = None
        self.fug_img = None
        self.house_img = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_gui()

    def _render_ansi(self):
        grid: UAVGrid = self.model.grid  # type: ignore
        grid_str = grid.get_ascii_repr(self._state[0], self._state[1])

        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [str(Direction(a)) for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")
        return "\n".join(output) + "\n"

    def _render_gui(self):
        try:
            import pygame
        except ImportError as e:
            raise DependencyNotInstalled(
                "pygame is not installed, run `pip install posggym[grid-world]`"
            ) from e

        if self.window_surface is None:
            pygame.init()

            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption("UAV")
                self.window_surface = pygame.display.set_mode(self.window_size)
            elif self.render_mode == "rgb_array":
                self.window_surface = pygame.Surface(self.window_size)

        assert (
            self.window_surface is not None
        ), "Something went wrong with pygame. This should never happen."

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.uav_img is None:
            file_name = path.join(path.dirname(__file__), "img/drone.png")
            self.uav_img = pygame.transform.scale(
                pygame.image.load(file_name), self.cell_size
            )
        if self.fug_img is None:
            file_name = path.join(path.dirname(__file__), "img/robber.png")
            self.fug_img = pygame.transform.scale(
                pygame.image.load(file_name), self.cell_size
            )
        if self.house_img is None:
            file_name = path.join(path.dirname(__file__), "img/house.png")
            self.house_img = pygame.transform.scale(
                pygame.image.load(file_name), self.cell_size
            )

        self.window_surface.fill((255, 255, 255))

        grid: UAVGrid = self.model.grid  # type: ignore
        uav_coord, fug_coord = self._state
        for y in range(grid.height):
            for x in range(grid.width):
                pos = (x * self.cell_size[0], y * self.cell_size[1])
                rect = (*pos, *self.cell_size)

                if (x, y) == grid.safe_house_coord:
                    self.window_surface.blit(self.house_img, pos)
                if (x, y) == fug_coord:
                    self.window_surface.blit(self.fug_img, pos)
                if (x, y) == uav_coord:
                    self.window_surface.blit(self.uav_img, pos)

                pygame.draw.rect(self.window_surface, (0, 0, 0), rect, 1)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
        elif self.render_mode == "rgb_array":
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
            )

    def close(self):
        if self.window_surface is not None:
            import pygame
            pygame.display.quit()
            pygame.quit()


class UAVModel(M.POSGModel[UAVState, UAVObs, UAVAction]):
    """Unmanned Aerial Vehicle Problem Model."""

    NUM_AGENTS = 2

    UAV_IDX = 0
    FUG_IDX = 1

    R_ACTION = -0.04
    R_CAPTURE = 1.0  # UAV reward, fugitive = -R_CAPTURE
    R_SAFE = -1.0  # UAV reward, fugitive = -R_SAFE

    # Observatio Accuracy for each agent
    FUG_OBS_ACC = 0.8
    UAV_OBS_ACC = 0.9

    def __init__(self, grid_name: str, **kwargs):
        self.grid = load_grid(grid_name)
        self._rng = random.Random(None)

        self.possible_agents = tuple(range(self.NUM_AGENTS))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        spaces.Discrete(self.grid.width),
                        spaces.Discrete(self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        coord_space = spaces.Tuple(
            (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
        )
        self.observation_spaces = {
            # UAV Obs = (uav coords, fug coords)
            self.UAV_IDX: spaces.Tuple((coord_space, coord_space)),
            # FUG ubs = (safe house dir)
            self.FUG_IDX: spaces.Discrete(len(DIR_OBS)),
        }
        self.observation_first = True
        self.is_symmetric = False

        # cache for sampling obs conditioned init state for fug
        self._cached_init_fug_obs: Optional[UAVFUGObs] = None
        self._valid_fug_coords_dist: Tuple[List[Coord], List[float]] = ([], [])

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        return {i: (self.R_SAFE, self.R_CAPTURE) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: UAVState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> UAVState:
        uav_coord = self.rng.choice(self.grid.init_uav_coords)
        fug_start_coords = list(self.grid.init_fug_coords)
        if uav_coord in fug_start_coords:
            fug_start_coords.remove(uav_coord)
        fug_coord = self.rng.choice(fug_start_coords)
        return uav_coord, fug_coord

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: UAVObs) -> UAVState:
        if agent_id == self.UAV_IDX:
            assert isinstance(obs, tuple)
            return self._sample_uav_obs(obs)  # type: ignore

        assert isinstance(obs, int)
        house_adj_coords = self.grid.get_neighbours(
            self.grid.safe_house_coord, ignore_blocks=False
        )
        if len(house_adj_coords) != 4:
            # Doesn't work for 3x3 grid
            raise NotImplementedError(
                "Sampling observation conditioned initial state for the fugitive is "
                "only supported for maps where safe house can be reached from all "
                "directions."
            )
        return self._sample_fug(obs)

    def _sample_uav_initial_state(self, uav_obs: UAVUAVObs) -> UAVState:
        uav_coord, fug_coord = uav_obs
        if self.rng.random() < self.UAV_OBS_ACC:
            # UAV_OBS_ACC probability obs loc is correct
            true_fug_coord = fug_coord
        else:
            fug_start_coords = list(self.grid.init_fug_coords)
            if uav_coord in fug_start_coords:
                fug_start_coords.remove(uav_coord)
            true_fug_coord = self.rng.choice(fug_start_coords)
        return uav_coord, true_fug_coord

    def _sample_fug(self, fug_obs: UAVFUGObs) -> UAVState:
        if self._cached_init_fug_obs is None or self._cached_init_fug_obs != fug_obs:
            self._valid_fug_coords_dist = self._get_fug_coord_dist(fug_obs)

        valid_coords, cum_probs = self._valid_fug_coords_dist
        fug_coord = self.rng.choices(  # type: ignore
            valid_coords, cum_weights=cum_probs, k=1
        )[0]
        uav_start_coords = list(self.grid.init_uav_coords)
        if fug_coord in uav_start_coords:
            uav_start_coords.remove(fug_coord)
        uav_coord = self.rng.choice(uav_start_coords)
        return uav_coord, fug_coord

    def _get_fug_coord_dist(self, obs: UAVFUGObs) -> Tuple[List[Coord], List[float]]:
        house_adj_coords = self.grid.get_neighbours(
            self.grid.safe_house_coord, ignore_blocks=False
        )

        if obs == OBSNONE:
            valid_fug_coords = set(self.grid.init_fug_coords)
            valid_fug_coords.difference_update(house_adj_coords)
            p = 1.0 / len(valid_fug_coords)
            dist = []
            for i in range(len(valid_fug_coords)):
                if i == 0:
                    dist.append(p)
                else:
                    dist.append(dist[-1] + p)
            return list(valid_fug_coords), dist
        if obs == OBSNORTH:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 1)
            true_coords = [house_adj_coords[Direction.SOUTH]]
        elif obs == OBSSOUTH:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 1)
            true_coords = [house_adj_coords[Direction.NORTH]]
        elif obs == OBSLEVEL:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 2)
            true_coords = [
                house_adj_coords[Direction.EAST],
                house_adj_coords[Direction.WEST],
            ]

        dist = []
        num_adj, num_true = len(house_adj_coords), len(true_coords)
        for coord in house_adj_coords:
            if coord in true_coords:
                dist.append(self.FUG_OBS_ACC / num_true)
            else:
                dist.append((1.0 - self.FUG_OBS_ACC) / (num_adj - num_true))
        return house_adj_coords, dist

    def sample_initial_obs(self, state: UAVState) -> Dict[M.AgentID, UAVObs]:
        return self._sample_obs(state)

    def step(
        self, state: UAVState, actions: Dict[M.AgentID, UAVAction]
    ) -> M.JointTimestep[UAVState, UAVObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        rewards = self._get_reward(next_state)

        uav_coord, fug_coord = next_state
        if fug_coord in (uav_coord, self.grid.safe_house_coord):
            # fug caught or fug safe
            fug_coord = self._sample_fug_coord(uav_coord)
            next_state = (uav_coord, fug_coord)

        obs = self._sample_obs(next_state)
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: UAVState, actions: Dict[M.AgentID, UAVAction]
    ) -> UAVState:
        uav_a, fug_a = actions[self.UAV_IDX], actions[self.FUG_IDX]
        uav_coord, fug_coord = state
        uav_next_coord = self.grid.get_next_coord(uav_coord, Direction(uav_a))
        # fugitive reseting is handled in step function
        # to allow for
        if uav_next_coord == fug_coord:
            # UAV considered to capture Fugitive
            fug_next_coord = fug_coord
        else:
            fug_next_coord = self.grid.get_next_coord(fug_coord, Direction(fug_a))
        return uav_next_coord, fug_next_coord

    def _sample_fug_coord(self, uav_coord: Coord) -> Coord:
        fug_start_coords = list(self.grid.init_fug_coords)
        if uav_coord in fug_start_coords:
            fug_start_coords.remove(uav_coord)
        return self.rng.choice(fug_start_coords)

    def _sample_obs(self, state: UAVState) -> Dict[M.AgentID, UAVObs]:
        return {
            self.UAV_IDX: self._sample_uav_obs(state),
            self.FUG_IDX: self._sample_fug_obs(state),
        }

    def _sample_uav_obs(self, state: UAVState) -> UAVUAVObs:
        uav_coord, fug_coord = state
        if (
            fug_coord == self.grid.safe_house_coord
            or self.rng.random() < self.UAV_OBS_ACC
        ):
            fug_coord_obs = fug_coord
        else:
            adj_coords = self.grid.get_neighbours(fug_coord)
            if uav_coord in adj_coords:
                adj_coords.remove(uav_coord)
            fug_coord_obs = self.rng.choice(adj_coords)
        return uav_coord, fug_coord_obs

    def _sample_fug_obs(self, state: UAVState) -> UAVFUGObs:
        fug_coord = state[1]
        fug_adj_coords = self.grid.get_neighbours(
            fug_coord, ignore_blocks=True, include_out_of_bounds=True
        )

        safe_house_coord = self.grid.safe_house_coord
        if fug_adj_coords[0] == safe_house_coord:
            true_obs = OBSSOUTH
        elif fug_adj_coords[1] == safe_house_coord:
            true_obs = OBSNORTH
        elif safe_house_coord in fug_adj_coords[2:]:
            true_obs = OBSLEVEL
        else:
            true_obs = OBSNONE

        if true_obs == OBSNONE:
            return true_obs

        if self.rng.random() < self.FUG_OBS_ACC:
            return true_obs
        return self.rng.choice([OBSNORTH, OBSSOUTH, OBSLEVEL])

    def _get_reward(self, next_state: UAVState) -> Dict[M.AgentID, SupportsFloat]:
        uav_coord, fug_coord = next_state
        rewards: Dict[M.AgentID, SupportsFloat] = {
            self.UAV_IDX: self.R_ACTION,
            self.FUG_IDX: self.R_ACTION,
        }
        if fug_coord == self.grid.safe_house_coord:
            rewards[self.UAV_IDX] = self.R_SAFE
            rewards[self.FUG_IDX] = -self.R_SAFE
        elif fug_coord == uav_coord:
            rewards[self.UAV_IDX] = self.R_CAPTURE
            rewards[self.FUG_IDX] = -self.R_CAPTURE
        return rewards


class UAVGrid(Grid):
    """A grid for the MA UAV Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Optional[Set[Coord]],
        safe_house_coord: Coord,
        init_fug_coords: List[Coord],
        init_uav_coords: List[Coord],
    ):
        super().__init__(grid_width, grid_height, block_coords)
        self.safe_house_coord = safe_house_coord
        self.init_fug_coords = init_fug_coords
        self.init_uav_coords = init_uav_coords

        self.valid_coords = set(self.unblocked_coords)
        self.valid_coords.remove(self.safe_house_coord)

    def get_ascii_repr(
        self, fug_coord: Optional[Coord], uav_coord: Optional[Coord]
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord == self.safe_house_coord:
                    row_repr.append("S")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if fug_coord is not None:
            grid_repr[fug_coord[0]][fug_coord[1]] = "F"
        if uav_coord is not None:
            if uav_coord == fug_coord:
                grid_repr[uav_coord[0]][uav_coord[1]] = "X"
            else:
                grid_repr[uav_coord[0]][uav_coord[1]] = "U"

        return (
            str(self) + "\n" + "\n".join(list(list((" ".join(r) for r in grid_repr))))
        )


def _empty_uav_grid(width, height, safe_house_coord) -> UAVGrid:
    all_coords = set(
        itertools.product(
            range(
                width,
            ),
            range(height),
        )
    )
    all_coords.remove(safe_house_coord)

    init_fug_coords = list(all_coords)
    init_uav_coords = [*all_coords]

    return UAVGrid(
        grid_height=width,
        grid_width=height,
        block_coords=None,
        safe_house_coord=safe_house_coord,
        init_fug_coords=init_fug_coords,
        init_uav_coords=init_uav_coords,
    )


def get_3x3_grid() -> UAVGrid:
    """Generate UAV 3x3 grid layout.

    .S.
    ...
    ...

    FUG and UAV can start at any point on grid that is not the safe house.

    |S| = 81
    |A_uav| = |A_fug| = 4
    |O_uav| = 81
    |O_fug| = 4
    """
    return _empty_uav_grid(3, 3, (1, 0))


def get_4x4_grid() -> UAVGrid:
    """Generate UAV 4x4 grid layout.

    ....
    ..S.
    ....
    ....

    FUG and UAV can start at any point on grid that is not the safe house

    |S| = 256
    |A_uav| = |A_fug| = 4
    |O_uav| = 256
    |O_fug| = 4
    """
    return _empty_uav_grid(4, 4, (2, 1))


def get_5x5_grid() -> UAVGrid:
    """Generate UAV 5x5 grid layout.

    .....
    ..S..
    .....
    .....
    .....

    FUG and UAV can start at any point on grid that is not the safe house

    |S| = 625
    |A_uav| = |A_fug| = 4
    |O_uav| = 625
    |O_fug| = 4
    """
    return _empty_uav_grid(5, 5, (2, 1))


def get_6x6_grid() -> UAVGrid:
    """Generate UAV 6x6 grid layout.

    ......
    ......
    ..S...
    ......
    ......
    ......

    FUG and UAV can start at any point on grid that is not the safe house

    |S| = 1296
    |A_uav| = |A_fug| = 4
    |O_uav| = 1296
    |O_fug| = 4

    """
    return _empty_uav_grid(6, 6, (2, 2))


SUPPORTED_GRIDS = {
    "3x3": get_3x3_grid,
    "4x4": get_4x4_grid,
    "5x5": get_5x5_grid,
    "6x6": get_6x6_grid,
}


def load_grid(grid_name: str) -> UAVGrid:
    """Load grid with given name."""
    grid_name = grid_name
    assert grid_name in SUPPORTED_GRIDS, (
        f"Unsupported grid name '{grid_name}'. Grid name must be one of: "
        f"{SUPPORTED_GRIDS.keys()}."
    )
    return SUPPORTED_GRIDS[grid_name]()



import sys
from typing import Tuple, Set

from gym import spaces

from posggym import Env

try:
    from ray.rllib.env.multi_agent_env import MultiAgentEnv
    from ray.rllib.utils.typing import MultiAgentDict, AgentID
except ImportError:
    print(
        "The posggym.wrapper.rllib_multi_agent_env Wrapper depends on the "
        "ray RLlib library. For installation instructions visit "
        "https://docs.ray.io/en/latest/rllib/index.html"
    )
    sys.exit(1)


class RllibMultiAgentEnv(MultiAgentEnv):
    """An interface between a POSGGym env and a Rllib MultiAgentEnv.

    This involves:
    - treating agent IDs as strings (instead of ints)
    - converting action, observations, and rewards from an ordered tuple to a
      dictionary mapping agent ID to the agent's action/observation/reward
    """

    def __init__(self, env: Env):
        # must assign this first before calling super().__init__() so that
        # property functions are initialized before super().__init__() is
        # called
        self.env = env
        self._agent_ids = set(str(i) for i in range(self.env.n_agents))
        super().__init__()

    @property
    def observation_space(self):
        """Get the environment observation space.

        This is a dictionary mapping agent IDs to the agent's observation space
        """
        return spaces.Dict({
            str(i): self.env.observation_spaces[i]
            for i in range(self.env.n_agents)
        })

    @property
    def action_space(self):
        """Get the environment action space.

        This is a dictionary mapping agent IDs to the agent's action space
        """
        return spaces.Dict(
            {
                str(i): self.env.action_spaces[i]
                for i in range(self.env.n_agents)
            }
        )

    @property
    def n_agents(self) -> int:
        return self.env.n_agents

    def get_agent_ids(self) -> Set[AgentID]:
        """Return a set of agent ids in the environment."""
        return self._agent_ids

    def reset(self) -> MultiAgentDict:
        obs = self.env.reset()
        return {str(i): o for i, o in enumerate(obs)}

    def step(self,
             action_dict: MultiAgentDict
             ) -> Tuple[
                 MultiAgentDict,
                 MultiAgentDict,
                 MultiAgentDict,
                 MultiAgentDict
             ]:
        actions = tuple(action_dict[str(i)] for i in range(self.env.n_agents))

        obs, rewards, done, info = self.env.step(actions)

        obs_dict = {str(i): o for i, o in enumerate(obs)}
        reward_dict = {str(i): r for i, r in enumerate(rewards)}
        done_dict = {str(i): done for i in range(self.env.n_agents)}
        info_dict = {str(i): info for i in range(self.env.n_agents)}

        # need to add RLlib specific special key to signify episode termination
        done_dict["__all__"] = done

        return obs_dict, reward_dict, done_dict, info_dict

    def render(self, mode=None):
        return_val = self.env.render(mode)
        if return_val is None:
            return True
        return return_val

    def close(self):
        self.env.close()

    @property
    def unwrapped(self):
        return self



"""The Multi-Access Broadcast problem.

A cooperative game involving control of a multi-access broadcast channel. In this
problem, each agent controls a node. Each node need to broadcast messages to each other
over a shared channel, with only one node able to broadcast at a time. If more than one
node broadcasts at the same time then there is a collision and no message is broadcast.
The nodes share the common goal of maximizing the throughput of the channel.

References
----------
- Ooi, J. M., and Wornell, G. W. 1996. Decentralized control of a multiple
  access broadcast channel: Performance bounds. In Proceedings of the 35th
  Conference on Decision and Control, 293–298.
- Hansen, Eric A., Daniel S. Bernstein, and Shlomo Zilberstein. “Dynamic
  Programming for Partially Observable Stochastic Games.” In Proceedings of
  the 19th National Conference on Artifical Intelligence, 709–715. AAAI’04.
  San Jose, California: AAAI Press, 2004.

"""
import sys
from itertools import product
from typing import Dict, List, Optional, SupportsFloat, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym.core import DefaultEnv
from posggym.utils import seeding


MABCState = Tuple[int, ...]
EMPTY = 0
FULL = 1
NODE_STATES = [EMPTY, FULL]
NODE_STATE_STR = ["E", "F"]

MABCAction = int
SEND = 0
NOSEND = 1
ACTIONS = [SEND, NOSEND]
ACTION_STR = ["S", "NS"]

MABCObs = int
COLLISION = 0
NOCOLLISION = 1
OBS = [COLLISION, NOCOLLISION]
OBS_STR = ["C", "NC"]


class MABCEnv(DefaultEnv[MABCState, MABCObs, MABCAction]):
    """The Multi-Access Broadcast Channel Environment.

    A cooperative game involving control of a multi-access broadcast channel.
    In this problem, each agent controls a node. Each node need to broadcast
    messages to each other over a shared channel, with only one node able to
    broadcast at a time. If more than one node broadcasts at the same time then
    there is a collision and no message is broadcast. The nodes share the
    common goal of maximizing the throughput of the channel.

    State
    -----
    Each node has a message buffer that can store up to one message at a time.
    That is it can be either `EMPTY` or `FULL`.

    Actions
    -------
    At each timestep each agent can either `SEND` a message or not (`NOSEND`).

    Observation
    -----------
    At the end of each time step, each node recieves a noisy observation of
    whether there was a `COLLISION` or `NOCOLLISION`.

    Each agent observes the true outcome with probability `obs_prob`, which is
    0.9 in the default version.

    Reward
    ------
    Each agent recieves a reward of `1` when a message is successfully
    broadcast and a reward of `0` otherwise.

    Transition Dynamics
    -------------------
    If a node's buffer is EMPTY then at each step it will become full with
    probability `fill_probs[i]`, otherwise it will remain empty (independent
    of the action performed).

    If a node's buffer is `FULL` then if the node does not send a message
    - i.e. uses the `NOSEND` action - then the buffer remains `FULL`.
    Otherwise - i.e. the agent chooses the `SEND` action - if no other nodes
    sends a message (there is no COLLISION) the buffer will be FULL with
    probability `fill_probs[i]`, otherwise it will be empty. If another node
    did sent a message as well then there was a COLLISION and the node's buffer
    remains `FULL`.

    References
    ----------
    - Ooi, J. M., and Wornell, G. W. 1996. Decentralized control of a multiple
      access broadcast channel: Performance bounds. In Proceedings of the 35th
      Conference on Decision and Control, 293–298.
    - Hansen, Eric A., Daniel S. Bernstein, and Shlomo Zilberstein. “Dynamic
      Programming for Partially Observable Stochastic Games.” In Proceedings of
      the 19th National Conference on Artifical Intelligence, 709–715. AAAI’04.
      San Jose, California: AAAI Press, 2004.

    """

    metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}

    def __init__(
        self,
        num_nodes: int = 2,
        fill_probs: Optional[Tuple[float, ...]] = None,
        observation_prob: float = 0.9,
        init_buffer_dist: Optional[Tuple[float, ...]] = None,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            MABCModel(
                num_nodes, fill_probs, observation_prob, init_buffer_dist, **kwargs
            ),
            render_mode=render_mode,
        )

    def render(self):
        assert self._last_obs is not None

        state_str = ", ".join([NODE_STATE_STR[s] for s in self._state])
        obs_str = ", ".join([OBS_STR[o] for o in self._last_obs.values()])
        output = [
            f"Step: {self._step_num}",
            f"State: <{state_str}>",
            f"Obs: <{obs_str}>",
        ]
        if self._last_actions is not None:
            action_str = ", ".join([ACTION_STR[a] for a in self._last_actions.values()])
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{tuple(self._last_rewards.values())}>")

        output_str = "\n".join(output) + "\n"

        if self.render_mode == "human":
            sys.stdout.write(output_str)
        else:
            # ansi mode
            return output_str


class MABCModel(M.POSGFullModel[MABCState, MABCObs, MABCAction]):
    """POSG Model for the Multi-Access Broadcast Channel problem."""

    DEFAULT_FILL_PROBS = (0.9, 0.1)
    DEFAULT_OBS_PROB = 0.9
    DEFAULT_INIT_BUFFER_DIST = (1.0, 1.0)

    R_SEND = 1.0
    R_NO_SEND = 0.0

    def __init__(
        self,
        num_nodes: int = 2,
        fill_probs: Optional[Tuple[float, ...]] = None,
        observation_prob: float = 0.9,
        init_buffer_dist: Optional[Tuple[float, ...]] = None,
        **kwargs,
    ):
        assert num_nodes >= 2

        if fill_probs is None:
            fill_probs = self.DEFAULT_FILL_PROBS
        if init_buffer_dist is None:
            init_buffer_dist = self.DEFAULT_INIT_BUFFER_DIST

        assert len(fill_probs) == num_nodes
        assert all(0 <= x <= 1 for x in fill_probs)
        assert 0 <= observation_prob <= 1
        assert len(init_buffer_dist) == num_nodes
        assert all(0 <= x <= 1 for x in init_buffer_dist)

        self._fill_probs = fill_probs
        self._obs_prob = observation_prob
        self._init_buffer_dist = init_buffer_dist

        self.possible_agents = tuple(range(num_nodes))
        self.state_space = spaces.Tuple(
            tuple(spaces.Discrete(len(NODE_STATES)) for i in self.possible_agents)
        )
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Discrete(len(OBS)) for i in self.possible_agents
        }
        self.observation_first = False
        self.is_symmetric = True

        # Spaces used internally
        self._state_space = list(
            product(*[list(NODE_STATES) for i in self.possible_agents])
        )
        self._action_spaces = tuple([*ACTIONS] for i in self.possible_agents)
        self._observation_spaces = tuple([*OBS] for i in self.possible_agents)

        self._trans_map = self._construct_trans_func()
        self._rew_map = self._construct_rew_func()
        self._obs_map = self._construct_obs_func()

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        return {i: (self.R_NO_SEND, self.R_SEND) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: MABCState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> MABCState:
        node_states = []
        for i in range(len(self.possible_agents)):
            if self.rng.random() <= self._init_buffer_dist[i]:
                node_states.append(FULL)
            else:
                node_states.append(EMPTY)
        return tuple(node_states)

    def step(
        self, state: MABCState, actions: Dict[M.AgentID, MABCAction]
    ) -> M.JointTimestep[MABCState, MABCObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        obs = self._sample_obs(actions)
        agent_reward = float(self._message_sent(state, actions)) * self.R_SEND
        rewards: Dict[M.AgentID, SupportsFloat] = {
            i: agent_reward for i in self.possible_agents
        }
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: MABCState, actions: Dict[M.AgentID, MABCAction]
    ) -> MABCState:
        next_node_states = list(state)
        for i, a_i in enumerate(actions):
            if a_i == SEND:
                # buffer emptied even if there is a collision
                next_node_states[i] = EMPTY
            if self.rng.random() <= self._fill_probs[i]:
                next_node_states[i] = FULL
        return tuple(next_node_states)

    def _sample_obs(
        self, actions: Dict[M.AgentID, MABCAction]
    ) -> Dict[M.AgentID, MABCObs]:
        senders = sum(int(a_i == SEND) for a_i in actions)
        if senders > 1:
            correct_obs = COLLISION
            wrong_obs = NOCOLLISION
        else:
            correct_obs = NOCOLLISION
            wrong_obs = COLLISION

        obs = {}
        for i in self.possible_agents:
            if self.rng.random() <= self._obs_prob:
                obs[i] = correct_obs
            else:
                obs[i] = wrong_obs
        return obs

    def get_initial_belief(self) -> Dict[MABCState, float]:
        b_map: Dict[MABCState, float] = {}
        s_prob_sum = 0.0
        for s in self._state_space:
            s_prob = 1.0
            for i in range(len(self.possible_agents)):
                if s[i] == FULL:
                    s_prob *= self._init_buffer_dist[i]
                else:
                    s_prob *= 1 - self._init_buffer_dist[i]
            b_map[s] = s_prob
            s_prob_sum += s_prob

        for s in self._state_space:
            b_map[s] /= s_prob_sum

        return b_map

    def transition_fn(
        self,
        state: MABCState,
        actions: Dict[M.AgentID, MABCAction],
        next_state: MABCState,
    ) -> float:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._trans_map[(state, action_tuple, next_state)]

    def _construct_trans_func(self) -> Dict:
        trans_map = {}
        agent_ids = [int(i) for i in self.possible_agents]
        for (s, a, s_next) in product(
            self._state_space, product(*self._action_spaces), self._state_space
        ):
            trans_prob = 1.0
            for i in agent_ids:
                if a[i] == NOSEND and s[i] == FULL:
                    if not s_next[i] == FULL:
                        trans_prob *= 0.0
                        break

                if s_next[i] == FULL:
                    trans_prob *= self._fill_probs[i]
                else:
                    trans_prob *= 1 - self._fill_probs[i]

            trans_map[(s, a, s_next)] = trans_prob
        return trans_map

    def observation_fn(
        self,
        obs: Dict[M.AgentID, MABCObs],
        next_state: MABCState,
        actions: Dict[M.AgentID, MABCAction],
    ) -> float:
        obs_tuple = tuple(obs[i] for i in self.possible_agents)
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._obs_map[(next_state, action_tuple, obs_tuple)]

    def _construct_obs_func(self) -> Dict:
        obs_map = {}
        agent_ids = [int(i) for i in self.possible_agents]
        for (s_next, a, o) in product(
            self._state_space,
            product(*self._action_spaces),
            product(*self._observation_spaces),
        ):
            senders = sum(int(a_i == SEND) for a_i in a)
            if senders > 1:
                correct_obs = COLLISION
            else:
                correct_obs = NOCOLLISION

            o_prob = 1.0
            for i in agent_ids:
                if o[i] == correct_obs:
                    o_prob *= self._obs_prob
                else:
                    o_prob *= 1 - self._obs_prob
            obs_map[(s_next, a, o)] = o_prob
        return obs_map

    def reward_fn(
        self, state: MABCState, actions: Dict[M.AgentID, MABCAction]
    ) -> Dict[M.AgentID, SupportsFloat]:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._rew_map[(state, action_tuple)]

    def _construct_rew_func(self) -> Dict:
        rew_map = {}
        joint_actions_space = product(*self._action_spaces)
        for (s, a) in product(self._state_space, joint_actions_space):
            reward = float(self._message_sent(s, a)) * self.R_SEND
            rew_map[(s, a)] = tuple(reward for _ in self.possible_agents)
        return rew_map

    def _message_sent(
        self,
        state: MABCState,
        actions: Union[Dict[M.AgentID, MABCAction], Tuple[MABCAction, ...]],
    ) -> bool:
        senders = sum(int(actions[int(i)] == SEND) for i in self.possible_agents)
        if senders != 1:
            return False

        message_sent = False
        for i in self.possible_agents:
            if actions[int(i)] == NOSEND:
                continue
            if state[int(i)] == FULL:
                message_sent = True
            break

        return message_sent



"""Model for the classic Multi-Agent Tiger problem.

This is a general-sum multi-agent version of the classic Tiger problem. It involves two
agents that are in a corridor facing two doors: `left` and `right`. Behind one door lies
a hungry tiger and behind the other lies treasure, but the agents do not know the
position of either the tiger or the treasure.

References
----------
- Gmytrasiewicz, Piotr J., and Prashant Doshi. “A Framework for Sequential
Planning in Multi-Agent Settings.” Journal of Artificial Intelligence
Research 24 (2005): 49–79.

"""
import sys
from itertools import product
from typing import Dict, List, Optional, SupportsFloat, Tuple

from gymnasium import spaces

import posggym.model as M
from posggym.core import DefaultEnv
from posggym.utils import seeding


MATState = int
TLEFT = 0
TRIGHT = 1
STATES = [TLEFT, TRIGHT]
STATE_STRS = ["TL", "TR"]

MATAction = int
OPENLEFT = 0
OPENRIGHT = 1
LISTEN = 2
ACTIONS = [OPENLEFT, OPENRIGHT, LISTEN]
ACTION_STR = ["OL", "OR", "L"]

MATObs = Tuple[int, int]
GROWLLEFT = 0
GROWLRIGHT = 1
CREAKLEFT = 0
CREAKRIGHT = 1
SILENCE = 2
OBS_PARTS = [(GROWLLEFT, GROWLRIGHT), (CREAKLEFT, CREAKRIGHT, SILENCE)]
OBS_SPACE = [
    (GROWLLEFT, CREAKLEFT),
    (GROWLLEFT, CREAKRIGHT),
    (GROWLLEFT, SILENCE),
    (GROWLRIGHT, CREAKLEFT),
    (GROWLRIGHT, CREAKRIGHT),
    (GROWLRIGHT, SILENCE),
]
OBS_STR = [("GL", "GR"), ("CL", "CR", "S")]


class MultiAgentTigerEnv(DefaultEnv):
    """The Multi-Agent Tiger Environment.

    This is a general-sum multi-agent version of the classic Tiger problem.

    This scenario involves two agents that are in a corridor facing two doors:
    `left` and `right`. Behind one door lies a hungry tiger and behind the
    other lies treasure, but the agents do not know the position of either the
    tiger or the treasure.

    State
    -----
    The state is defined by which door the tiger is behind. `TLEFT` for tiger
    is behind the `left` door, and `TRIGHT` for the tiger is behind the `right`
    door.

    S = {`TLEFT`, `TRIGHT`}

    The initial state is uniformly distributed between the possible states.

    Actions
    -------
    Each agent can either open the left-hand door `OPENLEFT`, open the
    right-hand door `OPENRIGHT`, or listen for the presence of the tiger
    `LISTEN`.

    A_1 = A_2 = {`OPENLEFT`, `OPENRIGHT`, `LISTEN`}

    Observation
    -----------
    Agents recieve observations of the position tiger: `GROWLEFT` for tiger
    left, and `GROWLRIGHT` for tiger right. Additionally, they also observe
    if a door has creaked: `CREAKLEFT` for left door, `CREAKRIGHT` for right
    door, and `SILENCE` for silence.

    O_1 = O_2 = {
        (`GROWLEFT`, `CREAKLEFT`),
        (`GROWLEFT`, `CREAKRIGHT`),
        (`GROWLEFT`, `SILENCE`),
        (`GROWLRIGHT`, `CREAKLEFT`),
        (`GROWLRIGHT`, `CREAKRIGHT`),
        (`GROWLRIGHT`, `SILENCE`)
    }

    If an agent uses the `LISTEN` action they will perceive the correct
    current position of the tiger with probability observation_prob=0.85,
    independent of if the other agent opens a door or listens. Furthermore,
    they will perceive the correct door opening or not with probability
    creak_observation_prob=0.9.

    If an agent opens either door they will recieve an observation uniformly at
    random.

    Reward
    ------
    Each agent recieves rewards independent of the other agent.

    An agent recieves a reward of +10 for opening the door without a tiger
    behind it, -100 for opening the door with the tiger behind it, and -1 for
    performing the listening action.

    Although the game is general-sum, and not zero-sum, agents influence each
    other by the way they influence the state.

    Transition Dynamics
    -------------------
    The state is reset to `TLEFT` or `TRIGHT` with equal probability whenever
    either agent opens one of the doors. Otherwise - when both agents perform
    the `LISTEN` action - the state is unchanged.

    References
    ----------
    - Gmytrasiewicz, Piotr J., and Prashant Doshi. “A Framework for Sequential
    Planning in Multi-Agent Settings.” Journal of Artificial Intelligence
    Research 24 (2005): 49–79.

    """

    metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}

    def __init__(
        self,
        observation_prob: float = 0.85,
        creak_observation_prob: float = 0.9,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            MultiAgentTigerModel(observation_prob, creak_observation_prob, **kwargs),
            render_mode=render_mode,
        )

    def render(self):
        assert self._last_obs is not None

        state_str = STATE_STRS[self._state]
        obs_str = ", ".join(
            [str((OBS_STR[0][o[0]], OBS_STR[1][o[1]])) for o in self._last_obs.values()]
        )
        output = [
            f"Step: {self._step_num}",
            f"State: <{state_str}>",
            f"Obs: <{obs_str}>",
        ]
        if self._last_actions is not None:
            action_str = ", ".join([ACTION_STR[a] for a in self._last_actions.values()])
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{tuple(self._last_rewards.values())}>")

        output_str = "\n".join(output) + "\n"

        if self.render_mode == "human":
            sys.stdout.write(output_str)
        else:
            # ansi mode
            return output_str


class MultiAgentTigerModel(M.POSGFullModel[MATState, MATObs, MATAction]):
    """Multi-Agent Tiger problem Model."""

    NUM_AGENTS = 2

    OPEN_GOOD_R = 10.0
    OPEN_BAD_R = -100.0
    LISTEN_R = -1.0

    def __init__(
        self,
        observation_prob: float = 0.85,
        creak_observation_prob: float = 0.9,
        **kwargs,
    ):
        assert 0 <= observation_prob <= 1.0
        assert 0 <= creak_observation_prob <= 1.0
        self._obs_prob = observation_prob
        self._creak_obs_prob = creak_observation_prob

        self.possible_agents = tuple(range(self.NUM_AGENTS))
        self.state_space = spaces.Discrete(len(STATES))
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Tuple(
                (spaces.Discrete(len(OBS_PARTS[0])), spaces.Discrete(len(OBS_PARTS[1])))
            )
            for i in self.possible_agents
        }
        self.observation_first = False
        self.is_symmetric = True

        # Spaces used internally
        self._state_space = STATES
        self._action_spaces = tuple(
            [*ACTIONS] for _ in range(len(self.possible_agents))
        )
        self._obs_spaces = tuple(OBS_SPACE for _ in range(len(self.possible_agents)))

        self._trans_map = self._construct_trans_func()
        self._rew_map = self._construct_rew_func()
        self._obs_map = self._construct_obs_func()

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        return {i: (self.OPEN_BAD_R, self.OPEN_BAD_R) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: MATState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> MATState:
        return self.rng.choice(STATES)

    def step(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> M.JointTimestep[MATState, MATObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        obs = self._sample_obs(state, actions)
        rewards = self._get_reward(state, actions)
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> MATState:
        next_state = state
        if any(a != LISTEN for a in actions.values()):
            next_state = self.rng.choice(STATES)
        return next_state

    def _sample_obs(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, MATObs]:
        obs: Dict[M.AgentID, MATObs] = {}
        for i, a in actions.items():
            if a != LISTEN:
                obs[i] = self.rng.choice(OBS_SPACE)
                continue

            other_agent_id = (int(i) + 1) % len(self.possible_agents)
            other_agent_action = actions[other_agent_id]

            tiger_obs = self._sample_tiger_obs(state)
            creak_obs = self._sample_creak_obs(other_agent_action)
            obs[i] = (tiger_obs, creak_obs)
        return obs

    def _sample_tiger_obs(self, state: MATState) -> int:
        if self.rng.random() < self._obs_prob:
            # correct obs
            return GROWLLEFT if state == TLEFT else GROWLRIGHT
        return GROWLRIGHT if state == TLEFT else GROWLLEFT

    def _sample_creak_obs(self, a_j: MATAction) -> int:
        if self.rng.random() < self._creak_obs_prob:
            # correct obs
            if a_j == LISTEN:
                return SILENCE
            return CREAKLEFT if a_j == OPENLEFT else CREAKRIGHT

        if a_j == LISTEN:
            return self.rng.choice([CREAKLEFT, CREAKRIGHT])
        if a_j == OPENLEFT:
            return self.rng.choice([CREAKRIGHT, SILENCE])
        return self.rng.choice([CREAKRIGHT, SILENCE])

    def _get_reward(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, SupportsFloat]:
        rewards: Dict[M.AgentID, SupportsFloat] = {}
        for i, a in actions.items():
            if a == LISTEN:
                rewards[i] = self.LISTEN_R
            elif a == state:
                rewards[i] = self.OPEN_BAD_R
            else:
                rewards[i] = self.OPEN_GOOD_R
        return rewards

    def get_initial_belief(self) -> Dict[MATState, float]:
        b_map: Dict[MATState, float] = {}
        for s in STATES:
            s_prob = 1.0 / len(STATES)
            b_map[s] = s_prob
        return b_map

    def transition_fn(
        self, state: MATState, actions: Dict[M.AgentID, MATAction], next_state: MATState
    ) -> float:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._trans_map[(state, action_tuple, next_state)]

    def _construct_trans_func(self) -> Dict:
        trans_map = {}
        uniform_prob = 1.0 / len(STATES)
        for (s, a, s_next) in product(
            self._state_space, product(*self._action_spaces), self._state_space
        ):
            if any(a_i != LISTEN for a_i in a):
                p = uniform_prob
            else:
                p = float(s_next == s)
            trans_map[(s, a, s_next)] = p
        return trans_map

    def observation_fn(
        self,
        obs: Dict[M.AgentID, MATObs],
        next_state: MATState,
        actions: Dict[M.AgentID, MATAction],
    ) -> float:
        obs_tuple = tuple(obs[i] for i in self.possible_agents)
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._obs_map[(next_state, action_tuple, obs_tuple)]

    def _construct_obs_func(self) -> Dict:
        obs_func = {}
        uniform_o_prob = 1.0 / len(OBS_SPACE)
        for (s_next, a, o) in product(
            self._state_space, product(*self._action_spaces), product(*self._obs_spaces)
        ):
            o_prob = 1.0

            a_i, a_j = a
            o_i, o_j = o

            correct_pos = GROWLLEFT if s_next == TLEFT else GROWLRIGHT

            if (a_i, a_j) == (LISTEN, LISTEN):
                for o_k in (o_i, o_j):
                    if o_k[0] == correct_pos:
                        o_prob *= self._obs_prob
                    else:
                        o_prob *= 1.0 - self._obs_prob

                    if o_k[1] == SILENCE:
                        o_prob *= self._creak_obs_prob
                    else:
                        o_prob *= (1.0 - self._creak_obs_prob) / 2

            elif LISTEN in (a_i, a_j):
                if a_i == LISTEN:
                    a_k, o_k = a_j, o_i
                else:
                    a_k, o_k = a_i, o_j

                o_prob *= uniform_o_prob
                correct_creak = CREAKLEFT if a_k == OPENLEFT else CREAKRIGHT

                if o_k[0] == correct_pos:
                    o_prob *= self._obs_prob
                else:
                    o_prob *= 1.0 - self._obs_prob

                if o_k[1] == correct_creak:
                    o_prob *= self._creak_obs_prob
                else:
                    o_prob *= (1.0 - self._creak_obs_prob) / 2

            else:
                o_prob *= uniform_o_prob * uniform_o_prob

            obs_func[(s_next, a, o)] = o_prob

        return obs_func

    def reward_fn(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, SupportsFloat]:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._rew_map[(state, action_tuple)]

    def _construct_rew_func(self) -> Dict:
        rew_map = {}
        joint_actions_space = product(*self._action_spaces)
        for (s, a) in product(self._state_space, joint_actions_space):
            rew_map[(s, a)] = self._get_reward(s, dict(enumerate(a)))
        return rew_map



import os
from typing import Callable, Optional, Tuple, Dict

import posggym.model as M
from posggym import logger
from posggym.core import Env, Wrapper
from posggym.wrappers.monitoring.video_recorder import VideoRecorder


def capped_cubic_video_schedule(episode_id: int) -> bool:
    """Get cubic schedule."""
    if episode_id < 1000:
        return int(round(episode_id ** (1.0 / 3))) ** 3 == episode_id
    else:
        return episode_id % 1000 == 0


class RecordVideo(Wrapper):
    """Wrapper for recording videos of rollouts.

    It is based on the gym.wrappers.RecordVideo (version gym>0.22) wrapper.
    Adapted here to work with posggym's multiagent environment.

    https://github.com/openai/gym/blob/0.22.0/gym/wrappers/record_video.py

    """

    def __init__(self,
                 env: Env,
                 video_folder: str,
                 episode_trigger: Optional[Callable[[int], bool]] = None,
                 step_trigger: Optional[Callable[[int], bool]] = None,
                 video_length: int = 0,
                 name_prefix: str = "posggym-video"):
        super().__init__(env)

        if episode_trigger is None and step_trigger is None:
            episode_trigger = capped_cubic_video_schedule

        trigger_count = sum(
            x is not None for x in [episode_trigger, step_trigger]
        )
        assert trigger_count == 1, "Must specify exactly one trigger"

        self.episode_trigger = episode_trigger
        self.step_trigger = step_trigger
        self.video_recorder = None

        self.video_folder = os.path.abspath(video_folder)
        # Create output folder if needed
        if os.path.isdir(self.video_folder):
            logger.warn(
                f"Overwriting existing videos at {self.video_folder} folder "
                "(try specifying a different `video_folder` for the "
                "`RecordVideo` wrapper if this is not desired)"
            )
        os.makedirs(self.video_folder, exist_ok=True)

        self.name_prefix = name_prefix
        self.step_id = 0
        self.episode_id = 0
        self.video_length = video_length

        self.recording = False
        self.recorded_frames = 0
        self.is_vector_env = getattr(env, "is_vector_env", False)

    def reset(self, **kwargs) -> M.JointObservation:
        observations = super().reset(**kwargs)
        if not self.recording and self._video_enabled():
            self.start_video_recorder()
        return observations

    def _video_enabled(self):
        if self.step_trigger:
            return self.step_trigger(self.step_id)
        return self.episode_trigger(self.episode_id)

    def start_video_recorder(self):
        """Starts video recorder."""
        self.close_video_recorder()

        video_name = f"{self.name_prefix}-step-{self.step_id}"
        if self.episode_trigger:
            video_name = f"{self.name_prefix}-episode-{self.episode_id}"

        base_path = os.path.join(self.video_folder, video_name)
        self.video_recorder = VideoRecorder(
            env=self.env,
            base_path=base_path,
            metadata={"step_id": self.step_id, "episode_id": self.episode_id},
        )

        self.video_recorder.capture_frame()
        self.recorded_frames = 1
        self.recording = True

    def step(self,
             actions: Tuple[M.Action, ...]
             ) -> Tuple[M.JointObservation, M.JointReward, bool, Dict]:
        observations, rewards, dones, infos = super().step(actions)

        # increment steps and episodes
        self.step_id += 1
        if not self.is_vector_env:
            if dones:
                self.episode_id += 1
        elif dones[0]:
            self.episode_id += 1

        if self.recording:
            self.video_recorder.capture_frame()
            self.recorded_frames += 1
            if self.video_length > 0:
                if self.recorded_frames > self.video_length:
                    self.close_video_recorder()
            else:
                if not self.is_vector_env:
                    if dones:
                        self.close_video_recorder()
                elif dones[0]:
                    self.close_video_recorder()

        elif self._video_enabled():
            self.start_video_recorder()

        return observations, rewards, dones, infos

    def close_video_recorder(self):
        """Closes the video recorder if currently recording."""
        if self.recording:
            self.video_recorder.close()
        self.recording = False
        self.recorded_frames = 1

    def close(self):
        super().close()
        self.close_video_recorder()

    def __del__(self):
        self.close_video_recorder()



"""Registers the internal POSGGym environments.

Based on the Farama Foundation Gymnasium API:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/envs/__init__.py

"""
import math

# import warnings
from itertools import product

from posggym.envs.grid_world import (
    driving,
    predator_prey,
    pursuit_evasion,
    two_paths,
    uav,
)
from posggym.envs.registration import make, pprint_registry, register, registry, spec


# Classic
# -------------------------------------------

register(
    id="MultiAccessBroadcastChannel-v0", entry_point="posggym.envs.classic.mabc:MABCEnv"
)

register(
    id="MultiAgentTiger-v0", entry_point="posggym.envs.classic.tiger:MultiAgentTigerEnv"
)

register(
    id="RockPaperScissors-v0",
    entry_point="posggym.envs.classic.rock_paper_scissors:RockPaperScissorsEnv",
)


# Grid World
# -------------------------------------------

# Driving
for grid_name in driving.SUPPORTED_GRIDS:
    grid_fn, finite_steps = driving.SUPPORTED_GRIDS[grid_name]
    for n in range(2, grid_fn().supported_num_agents + 1, 2):
        env_name_prefix = f"Driving-{grid_name}-n{n}"
        register(
            id=f"{env_name_prefix}-v0",
            entry_point="posggym.envs.grid_world.driving:DrivingEnv",
            max_episode_steps=finite_steps,
            kwargs={
                "grid": grid_name,
                "num_agents": n,
                "obs_dim": (3, 1, 1),
                "obstacle_collisions": False,
                "infinite_horizon": False,
            },
        )

        register(
            id=f"{env_name_prefix}-Hard-v0",
            entry_point="posggym.envs.grid_world.driving:DrivingEnv",
            max_episode_steps=finite_steps,
            kwargs={
                "grid": grid_name,
                "num_agents": n,
                "obs_dim": (3, 1, 1),
                "obstacle_collisions": True,
                "infinite_horizon": False,
            },
        )


# Driving Gen
for grid_name in driving.SUPPORTED_GEN_PARAMS:
    env_params = driving.SUPPORTED_GEN_PARAMS[grid_name]
    gen_params, finite_steps = env_params  # type: ignore
    register(
        id=f"Driving-Gen-{grid_name}-v0",
        entry_point="posggym.envs.grid_world.driving:DrivingGenEnv",
        max_episode_steps=finite_steps,
        kwargs={
            "num_agents": 2,
            "obs_dim": (3, 1, 1),
            "obstacle_collisions": False,
            "n_grids": None,
            "generator_params": gen_params,  # type: ignore
            "infinite_horizon": False,
        },
    )


# Level-Based Foraging
# Based on github.com/semitable/lb-foraging/

size_and_food = [(5, 4), (10, 8), (20, 10)]
players = [2, 4]  # reduced

for (size, n_food), n_agents, coop, static_layout in product(
    size_and_food, players, [True, False], [True, False]
):
    if n_agents > math.ceil(size / 2) * 4 - 4:
        continue

    coop_str = "-coop" if coop else ""
    static_layout_str = "-static" if static_layout else ""
    suffix = f"{size}x{size}-n{n_agents}-f{n_food}{coop_str}{static_layout_str}-v2"
    register(
        id=f"LevelBasedForaging-{suffix}",
        entry_point="posggym.envs.grid_world.lbf:LBFEnv",
        max_episode_steps=50,
        kwargs={
            "num_agents": n_agents,
            "max_agent_level": 3,
            "field_size": (size, size),
            "max_food": n_food,
            "sight": 2,
            "force_coop": coop,
            "static_layout": static_layout,
            "normalize_reward": True,
            "observation_mode": "tuple",
            "penalty": 0.0,
        },
    )

    register(
        id=f"LevelBasedForagig-VectorObs-{suffix}",
        entry_point="posggym.envs.grid_world.lbf:LBFEnv",
        max_episode_steps=50,
        kwargs={
            "num_agents": n_agents,
            "max_agent_level": 3,
            "field_size": (size, size),
            "max_food": n_food,
            "sight": 2,
            "force_coop": coop,
            "static_layout": static_layout,
            "normalize_reward": True,
            "observation_mode": "vector",
            "penalty": 0.0,
        },
    )

    register(
        id=f"LevelBasedForaging-GridObs-{suffix}",
        entry_point="posggym.envs.grid_world.lbf:LBFEnv",
        max_episode_steps=50,
        kwargs={
            "num_agents": n_agents,
            "max_agent_level": 3,
            "field_size": (size, size),
            "max_food": n_food,
            "sight": 2,
            "force_coop": coop,
            "static_layout": static_layout,
            "normalize_reward": True,
            "observation_mode": "grid",
            "penalty": 0.0,
        },
    )


# Predator-Prey
prey_strength = [1, 2, 3, 4]  # min is 1, max is 4 (since 4 adjacent cells)
num_prey = [1, 2, 3, 4]  # could be more, depending on grid size
num_predators = [2, 3, 4]  # max is 8 for default grids

for grid_name, coop, strength, n_prey, n_pred in product(
    predator_prey.SUPPORTED_GRIDS, [True, False], prey_strength, num_prey, num_predators
):
    if strength > n_pred:
        continue

    _, finite_steps = predator_prey.SUPPORTED_GRIDS[grid_name]
    coop_str = "-coop" if coop else ""
    env_name_suffix = f"{grid_name}-P{n_pred}-p{n_prey}-s{strength}{coop_str}-v0"
    register(
        id=f"PredatorPrey-{env_name_suffix}",
        entry_point="posggym.envs.grid_world.predator_prey:PPEnv",
        max_episode_steps=finite_steps,
        kwargs={
            "grid": grid_name,
            "num_predators": n_pred,
            "num_prey": n_prey,
            "cooperative": coop,
            "prey_strength": strength,
            "obs_dim": 2,
        },
    )


# Pursuit-Evasion
for grid_name in pursuit_evasion.SUPPORTED_GRIDS:
    register(
        id=f"PursuitEvasion-{grid_name}-v0",
        entry_point=("posggym.envs.grid_world.pursuit_evasion:PursuitEvasionEnv"),
        max_episode_steps=pursuit_evasion.SUPPORTED_GRIDS[grid_name][1],
        kwargs={
            "grid_name": grid_name,
            "action_probs": 1.0,
            "max_obs_distance": 12,
            "normalize_reward": True,
            "use_progress_reward": True,
        },
    )


# Two Paths
for grid_name in two_paths.SUPPORTED_GRIDS:
    register(
        id=f"TwoPaths-{grid_name}-v0",
        entry_point="posggym.envs.grid_world.two_paths:TwoPathsEnv",
        max_episode_steps=two_paths.SUPPORTED_GRIDS[grid_name][1],
        kwargs={"grid_name": grid_name, "action_probs": 1.0, "infinite_horizon": False},
    )


# # Unmanned Aerial Vehicle (UAV)
for grid_name in uav.SUPPORTED_GRIDS:
    register(
        id=f"UAV-{grid_name}-v0",
        entry_point="posggym.envs.grid_world.uav:UAVEnv",
        kwargs={
            "grid_name": grid_name,
        },
    )
